#include <cstdlib>
#include <ctime>
#include <ctype.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string> 
#include <windows.h>

typedef union {
	unsigned int pixel;
	struct {
		unsigned char blue;
		unsigned char green;
		unsigned char red;
		unsigned char alpha;
	};
} RGSSRGBA;

typedef struct {
	DWORD flags;
	DWORD class_type;
	void(*dmark) (void*);
	void(*dfree) (void*);
	double *data; //red is index 1, green is index 2, blue 3, alpha 0
} RGSSCOLOR;

typedef struct {
	DWORD unk1;
	DWORD unk2;
	BITMAPINFOHEADER *infoheader;
	RGSSRGBA *firstRow;
	RGSSRGBA *lastRow;
} RGSSBMINFO;

typedef struct {
	DWORD unk1;
	DWORD unk2;
	RGSSBMINFO *bminfo;
} BITMAPSTRUCT;

typedef struct {
	DWORD flags;
	DWORD class_type;
	void(*dmark) (void*);
	void(*dfree) (void*);
	BITMAPSTRUCT *bm;
} RGSSBITMAP;

# define M_PI           3.14159265358979323846  /* pi */

//===============================================================================
// * RotateRGSSBitmap
//  - Rotates Bitmaps in Game Memory
//        src_object : Object ID of RPG::Bitmap (bmp.object_id)
//     canvas_object : Object ID of Canvas to be drawn to - RPG::Bitmap
//             angle : Angle to rotate the Bitmap to in Degrees
//           showMsg : send a 1 to this Arg for a MessageBox
//===============================================================================
// Definately a work in progress.

extern "C" _declspec (dllexport) int RotateRGSSBitmap(unsigned long src_object, unsigned long canvas_object, float *pAngle, int showMsg = 0)
{
  // Get the Pointers to both Bitmaps for the Light Source and Canvas object to draw on
	RGSSBMINFO *lightSrc = ((RGSSBITMAP*)(src_object << 1))->bm->bminfo;
	RGSSBMINFO *lightCanvas = ((RGSSBITMAP*)(canvas_object << 1))->bm->bminfo;
	// Quit processing if there is no Light Source RGSS Bitmap
	if (!lightSrc || !lightCanvas) return 0;
	// Get the angle value from packed pointer in Arg
	double angle = pAngle[0];
		// RGSS specific local variables used to transfer from Source to Canvas
	RGSSRGBA *sourceData, *canvasData;
	// Local Variables for determining rotation
	int x, y, widthSrc, heightSrc, widthCanvas, heightCanvas;
	// Width and Height of Light Source and Canvas RGSS Bitmaps
	widthSrc = lightSrc->infoheader->biWidth;
	heightSrc = lightSrc->infoheader->biHeight;
	widthCanvas = lightCanvas->infoheader->biWidth;
	heightCanvas = lightCanvas->infoheader->biHeight;
	// Use Center of Canvas as center of rotation
	float cX = (float)widthCanvas / 2;
	float cY = (float)heightCanvas / 2;

	// Change the angle in Degrees to Radians
	double radianAngle = (M_PI * angle) / 180;
	// Determine Sine and Cosine of the Angle
	double cosAngle = (double)cos(radianAngle);
	double sinAngle = (double)sin(radianAngle);
	// Bitmap RGBA starting at last Pixel which holds 4 values, RGBA
	sourceData = lightSrc->lastRow;
	canvasData = lightCanvas->lastRow;
	// Points to specific Pixel Data
	RGSSRGBA *sourceRGBA, *canvasRGBA;
	// Holds Position of X and Y to transfer the data to
	double posX, posY;

	// Iterate through the Source Pixels and draw them on to the Canvas
	for (y = 0; y < heightSrc; y++)
	{
		for (x = 0; x < widthSrc; x++)
		{
			// Determine position on the Canvas Bitmap to draw
			posX = x * cosAngle + y * sinAngle;
			posY = x * sinAngle - y * cosAngle;
			// Recenter (not working right)
			posX += cX;
			posY += cY;
			// I dont have a clue what this does, I just copied and pasted
			sourceRGBA = sourceData + y + (x * widthSrc);
			canvasRGBA = canvasData + (int)posY + ((int)posX * widthCanvas);
			// If new position on canvas is actually within the canvas
			if (posY >= 0 && posY <= heightCanvas && posX >= 0 && posX <= widthCanvas)
			{
				// Transfer Data from the Source Image to the Canvas
				canvasRGBA->red = sourceRGBA->red;
				canvasRGBA->green = sourceRGBA->green;
				canvasRGBA->blue = sourceRGBA->blue;
				canvasRGBA->alpha = sourceRGBA->alpha;
			}
		}
	}

	// Debugging to display data inside DLL
	if (showMsg == 1)
	{
		// Change the variable name from "angle" here for displaying data
		// Remember to push the SHIFT key to use this Message Box
		std::string myStr = std::to_string(angle);
		MessageBoxA(NULL, myStr.c_str(), "DLL Message", MB_OK);
	}

	// Return Successful Rotation
	return 1;
}